from collections.abc import Mapping

from .core import _reify, _unify, construction_sentinel


def unifiable(cls):
    """Register standard unify and reify operations on a class.

    This uses the type and __dict__ or __slots__ attributes to define the
    nature of the term.

    >>> class A(object):
    ...     def __init__(self, a, b):
    ...         self.a = a
    ...         self.b = b
    >>> unifiable(A)
    <class 'unification.more.A'>

    >>> x = var('x')
    >>> a = A(1, 2)
    >>> b = A(1, x)

    >>> unify(a, b, {})
    {~x: 2}
    """
    _unify.add((cls, cls, Mapping), _unify_object)
    _reify.add((cls, Mapping), _reify_object)

    return cls


def _reify_object(o, s):
    """Reify a Python object with a substitution.

    >>> class Foo(object):
    ...     def __init__(self, a, b):
    ...         self.a = a
    ...         self.b = b
    ...     def __str__(self):
    ...         return "Foo(%s, %s)"%(str(self.a), str(self.b))

    >>> x = var('x')
    >>> f = Foo(1, x)
    >>> print(f)
    Foo(1, ~x)
    >>> print(reify_object(f, {x: 2}))
    Foo(1, 2)
    """
    if hasattr(o, "__slots__"):
        return _reify_object_slots(o, s)
    else:
        return _reify_object_dict(o, s)


def _reify_object_dict(o, s):
    obj = type(o).__new__(type(o))

    d = yield _reify(o.__dict__, s)

    yield construction_sentinel

    if d == o.__dict__:
        yield o
    else:
        obj.__dict__.update(d)
        yield obj


def _reify_object_slots(o, s):
    attrs = [getattr(o, attr) for attr in o.__slots__]
    new_attrs = yield _reify(attrs, s)

    yield construction_sentinel

    if attrs == new_attrs:
        yield o
    else:
        newobj = object.__new__(type(o))
        for slot, attr in zip(o.__slots__, new_attrs):
            setattr(newobj, slot, attr)

        yield newobj


def _unify_object(u, v, s):
    """Unify two Python objects.

    Unifies their type and ``__dict__`` attributes

    >>> class Foo(object):
    ...     def __init__(self, a, b):
    ...         self.a = a
    ...         self.b = b
    ...     def __str__(self):
    ...         return "Foo(%s, %s)"%(str(self.a), str(self.b))

    >>> x = var('x')
    >>> f = Foo(1, x)
    >>> g = Foo(1, 2)
    >>> unify_object(f, g, {})
    {~x: 2}
    """
    if type(u) != type(v):
        yield False
        return

    if hasattr(u, "__slots__"):
        yield _unify(
            tuple(getattr(u, slot) for slot in u.__slots__),
            tuple(getattr(v, slot) for slot in v.__slots__),
            s,
        )
    else:
        yield _unify(u.__dict__, v.__dict__, s)



from ._version import get_versions
from .core import assoc, reify, unify
from .more import unifiable
from .variable import Var, isvar, var, variables, vars

__version__ = get_versions()["version"]
del get_versions



from collections.abc import Mapping, Set
from contextlib import suppress


def transitive_get(key, d):
    """Get a value for a dict key in a transitive fashion.

    >>> d = {1: 2, 2: 3, 3: 4}
    >>> d.get(1)
    2
    >>> transitive_get(1, d)
    4
    """
    with suppress(TypeError):
        while key in d:
            key = d[key]
    return key


def _toposort(edges):
    """Topologically sort a dictionary.

    Algorithm by Kahn [1] - O(nodes + vertices).

    inputs:
        edges - a dict of the form {a: {b, c}} where b and c depend on a
    outputs:
        L - an ordered list of nodes that satisfy the dependencies of edges

    >>> _toposort({1: (2, 3), 2: (3, )})
    [1, 2, 3]

    Closely follows the wikipedia page [2]

    [1] Kahn, Arthur B. (1962), "Topological sorting of large networks",
    Communications of the ACM
    [2] http://en.wikipedia.org/wiki/Toposort#Algorithms
    """
    incoming_edges = reverse_dict(edges)
    incoming_edges = dict((k, set(val)) for k, val in incoming_edges.items())
    S = set((v for v in edges if v not in incoming_edges))
    L = []

    while S:
        n = S.pop()
        L.append(n)
        for m in edges.get(n, ()):
            assert n in incoming_edges[m]
            incoming_edges[m].remove(n)
            if not incoming_edges[m]:
                S.add(m)
    if any(incoming_edges.get(v, None) for v in edges):
        raise ValueError("Input has cycles")
    return L


def reverse_dict(d):
    """Reverses the direction of a dependency dict.

    >>> d = {'a': (1, 2), 'b': (2, 3), 'c':()}
    >>> reverse_dict(d)  # doctest: +SKIP
    {1: ('a',), 2: ('a', 'b'), 3: ('b',)}

    :note: dict order are not deterministic. As we iterate on the
        input dict, it make the output of this function depend on the
        dict order. So this function output order should be considered
        as undeterministic.

    """
    result = {}
    for key in d:
        for val in d[key]:
            result[val] = result.get(val, tuple()) + (key,)
    return result


def freeze(d):
    """Freeze container to hashable a form.

    >>> freeze(1)
    1

    >>> freeze([1, 2])
    (1, 2)

    >>> freeze({1: 2}) # doctest: +SKIP
    ((1, 2),)
    """
    if isinstance(d, Mapping):
        return tuple(map(freeze, sorted(d.items(), key=lambda x: hash(x[0]))))
    if isinstance(d, Set):
        return tuple(map(freeze, sorted(d, key=hash)))
    if isinstance(d, (tuple, list)):
        return tuple(map(freeze, d))
    return d
